# date formatting and arithmetic

    package	Date_fns;
    require	Exporter;
    @ISA	= qw(Exporter);
    @EXPORT	= qw(
    date is_greater_equal_date isleapyear one_month_prior last_month
    one_month_hence first_next_month one_quarter_hence one_year_prior
    one_quarter_prior
    one_year_hence unpack_date pack_date is_less_than_date
    n_days_hence n_days_prior get_activity_month get_activity_date
    ninety_days_hence ninety_days_prior days_between compare_dates
    first_next_period first_last_period one_period_hence one_period_prior
    timestamp2date format_date first_next_month 
    first_this_month format_ts sub_time pack_time unpack_time
    add_time add_time_c sub_time_c Format_Time get_activity_time
    date2doy julian2date doy2date build_dow get_dow unpack_timestamp
    get_hour now date2dow is_holiday get_activity_year sec_2_hms
    );

=head1 NAME

date_fns - routines for manipulating dates.

=cut

    sub date () ;
    sub unpack_date ($) ;
    sub pack_date ($$;$) ;
    sub compare_dates ($$) ;
    sub is_greater_equal_date ($$) ;
    sub is_less_than_date ($$) ;
    sub days_between ($$) ;
    sub isleapyear ($) ;
    sub one_month_prior ($) ;
    sub last_month ($) ;
    sub one_month_hence ($) ;
    sub first_this_month ($) ;
    sub first_next_month ($) ;
    sub one_quarter_hence ($) ;
    sub one_quarter_prior ($) ;
    sub ninety_days_hence ($) ;
    sub ninety_days_prior ($) ;
    sub one_year_prior ($) ;
    sub one_year_hence ($) ;
    sub n_days_hence ($;$) ;
    sub n_days_prior ($;$) ;
    sub one_period_hence ($$) ;
    sub first_next_period ($$) ;
    sub one_period_prior ($$) ;
    sub last_next_period ($$) ;
    sub get_activity_month ($$) ;
    sub get_activity_year ($$) ;
    sub get_activity_date ($$;$) ;
    sub timestamp2date ($) ;
    sub format_date ($;$) ;
    sub add_time ($$) ;
    sub add_time_c ($$) ;
    sub sub_time ($$) ;
    sub sub_time_c ($$) ;
    sub pack_time ($$$;$) ;
    sub unpack_time ($) ;
    sub unpack_timestamp ($) ;
    sub Format_Time ($$$;$) ;
    sub get_activity_time ($$) ;
    sub date2doy ($) ;
    sub julian2date ($) ;
    sub doy2date ($) ;
    sub build_dow ($;$) ;
    sub is_holiday ($;$$$) ;
    sub get_dow () ;
    sub get_hour () ;
    sub now () ;
    sub date2dow ($);
    sub sec_2_hms (*);

    @daynames = ("Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat");
    @months=("Jan", "Feb", "Mar", "Apr", "May", "Jun",
	     "Jul", "Aug", "Sep", "Oct", "Nov", "Dec");
    @daypermonth=(31, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31);

    %month2int = (
        "January" => 1,
        "February" => 2,
        "March" => 3,
        "April" => 4,
        "May" => 5,
        "June" => 6,
        "July" => 7,
        "August" => 8,
        "September" => 9,
        "October" => 10,
        "November" => 11,
        "December" => 12,
    );

    1;

=head1 date

=head2 DESCRIPTION

return date as yyyymmdd plus month (1-12), numerical day of month, year.
if fakedate exists, use it instead of today's date.

=head2 ARGUMENTS

=over 5

=item void

=back

=head2 RETURN VALUES

=over 5

=item array "yyyymmdd", "yyyy", "mm", "dd"

=back

=cut
sub date () {
my($dt,$m,$d,$y,$sec,$min,$hour);
    if($main::fakedate) {
	($y, $m, $d) = unpack_date($main::fakedate);
	return($main::fakedate,$m,$d,$y);
    }
    ($sec,$min,$hour,$d,$m,$y) = localtime time;
    $y += 1900;
    $m++;
    $dt = pack_date($y, $m, $d);
    return($dt,$m,$d,$y);
}

sub now () {
my($dt,$m,$d,$y,$sec,$min,$hour);
    ($sec,$min,$hour,$d,$m,$y) = localtime time;
    return($hour,$min,$sec);
}

=head1 unpack_date

=head2 DESCRIPTION

unpack date from yyyymmdd format to array of y,m,d

=head2 ARGUMENTS

=over 5

=item date string yyyymmdd

=back

=head2 RETURN VALUES

=over 5

=item array (yyyy, mm, dd)

=back

=cut
sub unpack_date ($) {
my($date) = @_;
    $date =~ s/-//g;
    return(unpack("a4a2a2", $date));
}

=head1 pack_date

=head2 DESCRIPTION

Takes date as array (year, month, day), returns yyyymmdd format.
Defaults to first of month if day not supplied.

=head2 ARGUMENTS

=over 5

=item integer year

=item integer month (1-12)

=item optional integer day (1-31)

=back

=head2 RETURN VALUES

=over 5

=item string "yyyymmdd"

=back

=cut
sub pack_date ($$;$) {
my($y,$m,$d) = @_;
    $d ||= 1;
    return(sprintf("%04d%02d%02d", $y, $m, $d));
}

=head1 compare_dates

=head2 DESCRIPTION

compare dates in fmt yyyymmdd and return comparison result

=head2 ARGUMENTS

=over 5

=item dta "yyyymmdd"

=item dtb "yyyymmdd"

=back

=head2 RETURN VALUES

=over 5

=item integer result of <=> comparison

=back

=cut
sub compare_dates ($$) {
    my($dta, $dtb) = @_;
    return($dta <=> $dtb);
}

=head1 is_greater_equal_date

=head2 DESCRIPTION

compare dates in fmt yyyymmdd and return comparison result

=head2 ARGUMENTS

=over 5

=item dta "yyyymmdd"

=item dtb "yyyymmdd"

=back

=head2 RETURN VALUES

=over 5

=item integer result of >= comparison

=back

=cut
sub is_greater_equal_date ($$) {
    my($dta, $dtb) = @_;
    return($dta >= $dtb);
}

=head1 is_less_than_date

=head2 DESCRIPTION

compare dates in fmt yyyymmdd and return comparison result

=head2 ARGUMENTS

=over 5

=item dta "yyyymmdd"

=item dtb "yyyymmdd"

=back

=head2 RETURN VALUES

=over 5

=item integer result of < comparison

=back

=cut
sub is_less_than_date ($$) {
    my($dta, $dtb) = @_;
    return($dta < $dtb);
}

=head1 days_between

=head2 DESCRIPTION

calculate days from dta to dtb (dtb - dta)

=head2 ARGUMENTS

=over 5

=item dta "yyyymmdd"

=item dtb "yyyymmdd"

=back

=head2 RETURN VALUES

=over 5

=item integer number of days from dta to dtb

=back

=cut
sub days_between ($$) {
    my($dta, $dtb) = @_;
    my($ma, $da, $ya, $mb, $db, $yb, $days, $m, $y, $sign, @daypermonth);
    @daypermonth=(31, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31);
    if($dta == $dtb) { return 0 ; }
    $sign = 1;
    if(is_greater_equal_date($dta, $dtb)) {
	($dtb, $dta) = ($dta, $dtb);
	$sign = -1;
    }
    if($main::debug>10) { print "DB0: ($dtb - $dta), sign = $sign\n"; }
    ($ya, $ma, $da) = unpack("a4a2a2", $dta);
    ($yb, $mb, $db) = unpack("a4a2a2", $dtb);
    $daypermonth[2] = (isleapyear($yb)) ? 29 : 28;
    if($ma > 12) { $ma = 12; }	# prevent infinite loop on bad data
    if($mb > 12) { $mb = 12; }	# prevent infinite loop on bad data
    $days = ($db - $da);
    if($main::debug>10) { print "DB1: days = ($db - $da) = $days, mb = $mb\n"; }
    if($days < 0) { 
        $mb--;
	if($mb < 1) {
	    $yb--;
	    $daypermonth[2] = (isleapyear($yb)) ? 29 : 28;
	    $mb = 12;
	    if($main::debug>10) { print "DB2: yb = $yb, mb = $mb\n"; }
	}
	if($main::debug>10) { print "DB3: mb = $mb, days = $days\n"; }
        $days += $daypermonth[$mb];
	if($main::debug>10) { print "DB4: mb = $mb, days = $days\n"; }
    }   
    while($mb != $ma) {
	if($main::debug>10) { print "DB5: adding " . $daypermonth[$ma] . " days for month $ma\n"; }
	$days += $daypermonth[$ma++];
	if($ma > 12) {
	    $ya++;
	    $ma = 1;
	    if($main::debug>10) { print "DB6: ya = $ya, ma = $ma\n"; }
	}
    }
    $y = ($yb - $ya);
    if($main::debug>10) { print "DB7: days = " . ($days + ($y * 365)) . "\n"; }
    return($sign * ($days + ($y * 365)));
}

=head1 isleapyear

=head2 DESCRIPTION

return 1 if year is a leap year

=head2 ARGUMENTS

=over 5

=item integer year as yyyy

=back

=head2 RETURN VALUES

=over 5

=item integer 1 if leap year, else 0

=back

=cut
sub isleapyear ($) {
    my ($year) = @_;
    if(($year+0) % 4) {
	$main::daypermonth[2] = 28;
	return(0);
    }
    if(0 == (($year+0) % 400)) {
	$main::daypermonth[2] = 29;
	return(1);
    }
    if(0 != (($year+0) % 100)) {
	$main::daypermonth[2] = 29;
	return(1);
    }
    $main::daypermonth[2] = 28;
    return(0);
}

=head1 one_month_prior

=head2 DESCRIPTION

takes date in yyyymmdd format, returns same day of previous month, same format

=head2 ARGUMENTS

=over 5

=item string "yyyymmdd"

=back

=head2 RETURN VALUES

=over 5

=item string "yyyymmdd"

=back

=cut
sub one_month_prior ($) {
my ($date) = @_;
my ($d,$m,$y);
    $date =~ s/[^\d]//g;
    if($date < 00000100) { return($date); }	# don't work B.C.E.
    ($y, $m, $d) = unpack("a4a2a2", $date);
    $m--;
    if($m < 1) {
	$m = 12;
	$y -= 1;
    }
    # must check if this is a legal date (not, for example, 2008 02 31)
    if(isleapyear($y)) {
	$daypermonth[2] = 29;
    }
    if($d > $daypermonth[$m]) { $d = $daypermonth[$m]; }
    $r = pack_date($y, $m, $d);
    return($r);
}

=head1 last_month

=head2 DESCRIPTION

takes date in yyyymmdd format, returns first of previous month in same format

=head2 ARGUMENTS

=over 5

=item string "yyyymmdd"

=back

=head2 RETURN VALUES

=over 5

=item string "yyyymmdd"

=back

=cut
sub last_month ($) {
my ($date) = @_;
    $date = first_this_month(one_month_prior($date));
    return($date);
}

=head1 one_month_hence

=head2 DESCRIPTION

takes date in yyyymmdd format, returns one month later in same format

=head2 ARGUMENTS

=over 5

=item string "yyyymmdd"

=back

=head2 RETURN VALUES

=over 5

=item string "yyyymmdd"

=back

=cut
sub one_month_hence ($) {
    my ($date) = @_;
    my ($d,$m,$y,$r,$p,$ly,@daypermonth);
    @daypermonth=(31, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31);
    unless($date) { return(""); }
    $date =~ s/[^\d]//g;
    ($y, $m, $d) = unpack("a4a2a2", $date);
    $m++;
    if($m > 12) {
	$m = 1;
	$y += 1;
    }
    if(isleapyear($y)) {
	$daypermonth[2] = 29;
    }
    if($d > $daypermonth[$m]) { $d = $daypermonth[$m]; }
    $r = pack_date($y, $m, $d);
    return($r);
}

#takes date in yyyymmdd format, returns first of this month in same format
sub first_this_month ($) {
my ($date) = @_;
    unless($date) { return(""); }
    $date =~ s/[^\d]//g;
    substr($date,6,2) = "01";
    return($date);
}

#takes date in yyyymmdd format, returns first of next month in same format
sub first_next_month ($) {
my ($date) = @_;
    unless($date) { return(""); }
    $date = first_this_month(one_month_hence($date));
    return($date);
}

# takes date in yyyymmdd format, returns same day of month 3 months later
# in same format
sub ninety_days_hence ($) {
    my ($date) = @_;
    my ($d,$m,$y,$edate,@daypermonth);
    @daypermonth=(31, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31);
    unless($date) { return(""); }
    $date =~ s/[^\d]//g;
    ($y, $m, $d) = unpack("a4a2a2", $date);
    $m += 3;
    if($m > 12) {
	$m -= 12;
	$y += 1;
    }
    if(isleapyear($y)) {
	$daypermonth[2] = 29;
    }
    if($d > $daypermonth[$m]) { $d = $daypermonth[$m]; }
    $edate = pack_date($y, $m, $d);
    return($edate);
}

# takes date in yyyymmdd format, returns same day of month 3 months earlier
# in same format
sub ninety_days_prior ($) {
    my ($date) = @_;
    my ($d,$m,$y,$edate,@daypermonth);
    @daypermonth=(31, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31);
    ($y, $m, $d) = unpack("a4a2a2", $date);
    $m -= 3;
    if($m < 1) {
	$m += 12;
	$y -= 1;
    }
    if(isleapyear($y)) {
	$daypermonth[2] = 29;
    }
    if($d > $daypermonth[$m]) { $d = $daypermonth[$m]; }
    $edate = pack_date($y, $m, $d);
    return($edate);
}

# takes date in yyyymmdd format, returns first of month three months later
# in same format
sub one_quarter_hence ($) {
my($date) = @_;
my($edate);
    $edate = first_next_month($date);
    $edate = first_next_month($edate);
    $edate = first_next_month($edate);
    return($edate);
}

# takes date in yyyymmdd format, returns first of month three months prior
# in same format
sub one_quarter_prior ($) {
my($date) = @_;
my($edate);
    $edate = last_month($date);
    $edate = last_month($edate);
    $edate = last_month($edate);
    return($edate);
}

#takes date in yyyymmdd format, returns date one year in past in same format
sub one_year_prior ($) {
my ($date) = @_;
my ($r, $y, $m, $d);
    ($y, $m, $d) = unpack("a4a2a2", $date);
    $y--;
    $r = pack_date($y, $m, $d);
    return($r);
}

#takes date in yyyymmdd format, returns date one year in future in same format
sub one_year_hence ($) {
my ($date) = @_;
my ($r, $y, $m, $d);
    ($y, $m, $d) = unpack("a4a2a2", $date);
    $y++;
    $r = pack_date($y, $m, $d);
    return($r);
}

# Takes date in yyyymmdd format, returns date 1 day in future in same format.
# If second argument, use that instead of 1 day.
sub n_days_hence ($;$) {
my ($date, $days) = @_;
my ($d,$df,$m,$y,$ly,$r,@daypermonth);
    @daypermonth=(31, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31);
    ($y, $m, $d) = unpack("a4a2a2", $date);
    $ly = isleapyear($y);
    $daypermonth[2] = ($ly) ? 29 : 28 ;
    $days ||= 1;
    $days or return($date);
    $df = $d + $days;
    $p = 0 + $daypermonth[$m+0];
    if($main::debug>11) { print "y='$y', m='$m', d='$d', p='$p'\n"; }
    while($df > $p) {
	if($main::debug>11) { print "too many days: y='$y', m='$m', df='$df' > p='$p'\n"; }
	$m++;
	if($m > 12) {
	    if($main::debug>11) { print "too many months: y='$y', m='$m', df='$df', p='$p'\n"; }
	    $m = 1;
	    $y++;
	    $ly = isleapyear($y);
	    $daypermonth[2] = ($ly) ? 29 : 28 ;
	}
	$df -= $p;
	$p = 0 + $daypermonth[$m+0];
    }
    $d = $df;
    $r = pack_date($y, $m, $d);
    return($r);
}

#takes date in yyyymmdd format, returns date 1 day in past in same format
# If second argument, use that instead of 1 day.
sub n_days_prior ($;$) {
my ($date, $days) = @_;
my ($d,$df,$m,$y,$ly,$r,@daypermonth);
    @daypermonth=(31, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31);
    # printf main::BIGDBG ("n_days_prior(%s, %s)\n", $date, $days);
    ($y, $m, $d) = unpack("a4a2a2", $date);
    # printf main::BIGDBG ("unpack to %s %s %s\n", $y, $m, $d);
    $ly = isleapyear($y);
    $daypermonth[2] = ($ly) ? 29 : 28 ;
    $days ||= 1;
    $days or return($date);
    $df = $d - $days;
    # printf main::BIGDBG ("subtract %d to get %s %s %s\n", $days, $y, $m, $df);
    while($df < 1) {
	$main::DXCdebug && print "too few days: y='$y', m='$m', df='$df'< 1, p='$p'\n";
	$m--;		# back up a month
	if($m < 1) {	# if less than January
	    $main::DXCdebug && print "too few months: y='$y', m='$m', df='$df', p='$p'\n";
	    $m = 12;	# month = December
	    $y--;	# prior year
	    $ly = isleapyear($y);
	    $daypermonth[2] = ($ly) ? 29 : 28 ;
	}
	$p = 0 + $daypermonth[$m+0];
	$main::DXCdebug && print "days for month $m = $p\n";
	$df += $p;
    }
    $d = $df;
    $r = pack_date($y, $m, $d);
    # printf main::BIGDBG ("pack %s %s %s to %s\n", $y, $m, $d, $r);
    return($r);
}

# return same day in month one period in future
sub one_period_hence ($$) {
my($date, $intvmos)=@_;
    if($intvmos == 1) {
	$date = one_month_hence($date);	# same day in month
    }
    elsif($intvmos == 3) {
	$date = ninety_days_hence($date);  # same day in month
    }
    elsif($intvmos >= 12) {
	$date = one_year_hence($date);	# same day in month
	if($intvmos == 24) {
	    $date = one_year_hence($date); # same day in month
	}
    }
    return($date);
}

# return same day in month one period in future, rounded down to first of month
sub first_next_period ($$) {
my($date, $intvmos)=@_;
    $date = first_this_month(one_period_hence($date,$intvmos));
    return($date);
}

sub one_period_prior ($$) {
my($date, $intvmos)=@_;
    if($intvmos == 1) {
	$date = one_month_prior($date); # same day of month
    }
    elsif($intvmos == 3) {
	$date = ninety_days_prior($date); # same day of month
    }
    elsif($intvmos >= 12) {
	$date = one_year_prior($date);	# same day of month
	if($intvmos == 24) {
	    $date = one_year_prior($date);	# same day of month
	}
    }
    return($date);
}

# return date one period in past, rounded down to 1st of month
sub first_last_period ($$) {
my($date, $intvmos)=@_;
    $date = first_this_month(one_period_prior($date,$intvmos));
    return($date);
}

sub get_activity_month ($$) {
my($ddate, $activity) = @_;
my($fdate, $omonth, $dmonth, $dday, $dyear);
    $ddate ||= "";
    if($ddate) {
	($dyear, $dmonth, $dday) = unpack("a4a2a2", $ddate);
    }
    else {
	($ddate, $dmonth, $dday, $dyear) = date();
    }
    $fdate = format_date($ddate, "MM yyyy");
    if(!cdb_fns::ok("Use this date ($fdate) for \u\L$activity month?")) {
	do {
	    $omonth = $dmonth;
	    do {
		$dmonth = cdb_fns::getv("$activity month (01 to 12):",$dmonth);
	    } until (cdb_fns::bounds($dmonth, 1, 12, "month"));
	    do {
		if($dyear < 1990) { print "Year must be later than 1990\n"; }
		if($dyear > 2100) { print "Year must be earlier than 2100\n"; }
		
		if($omonth == 12 && $dmonth == 1) { $dyear++; }
		if($omonth <=  3 && $dmonth >= 6) { $dyear--; }
		$dyear = cdb_fns::getv("$activity year", $dyear);
	    } until ($dyear > 1990 && $dyear < 2100);
	} until cdb_fns::ok();
    }
    $ddate = pack_date($dyear, $dmonth);
    $main::silent || cdb_fns::W() || print "Using $dmonth/$dyear for \u\L$activity month.\n";
    return($dmonth, $dyear, $ddate);
}

sub get_activity_year ($$) {
my($ddate, $activity) = @_;
    $ddate ||= "";
    if($ddate) {
	($dyear, $dmonth, $dday) = unpack("a4a2a2", $ddate);
    }
    else {
	($ddate, $dmonth, $dday, $dyear) = date();
    }
    $ddate = pack_date($dyear, 1);
    $fdate = format_date($ddate, "MM yyyy");
    if(!cdb_fns::ok("Use this date ($fdate) for \u\L$activity year?")) {
	do {
	    if($dyear < 1990) { print "Year must be later than 1990\n"; }
	    if($dyear > 2100) { print "Year must be earlier than 2100\n"; }
	    
	    if($omonth == 12 && $dmonth == 1) { $dyear++; }
	    if($omonth <=  3 && $dmonth >= 6) { $dyear--; }
	    $dyear = cdb_fns::getv("$activity year", $dyear);
	} until ($dyear > 1990 && $dyear < 2100);
    }
    $ddate = pack_date($dyear, 1);
    return(1, $dyear, $ddate);
}

sub get_activity_date ($$;$) {
my($ddate, $activity, $help) = @_;
my($fdate, $omonth, $dmonth, $dday, $dyear, $mdays, @daypermonth, $prompt);

    $help ||= "";
    @daypermonth=(31, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31);
    $ddate ||= "";
    if($ddate) {
	($dyear, $dmonth, $dday) = unpack("a4a2a2", $ddate);
    }
    else {
	($ddate, $dmonth, $dday, $dyear) = date();
    }
    $fdate = format_date($ddate, "MM dd, yyyy");
    if(!cdb_fns::ok("Use this date ($fdate) for \u\L$activity date?")) {
	($dyear, $dmonth, $dday) = unpack("a4a2a2", $ddate);
	$daypermonth[2] = (isleapyear($dyear)) ? 29 : 28 ;
	$mdays = $daypermonth[$dmonth];
	do {
	    $omonth = $dmonth;
	    do {
		$prompt = "$activity day";
		if($help) { $prompt .= " (? for help)"; }
		$dday=cdb_fns::getv($prompt, $dday);
		if(($dday eq "?"
		|| $dday =~ /^h/i)
		&& $help ne "") {
		    ok($help);
		    $dday = 0;
		}
	    } until (cdb_fns::bounds($dday, 1, 31, "day"));
	    do {
		$dmonth = cdb_fns::getv("$activity month (01 to 12):",$dmonth);
	    } until (cdb_fns::bounds($dmonth, 1, 12, "month"));
	    do {
		if($dyear < 1990) {
		    print "Year must be later than 1990\n";
		}
		if($dyear > 2100) {
		    print "Year must be earlier than 2100\n";
		}
		if($omonth == 12 && $dmonth == 1) { $dyear++; }
		if($omonth <=  3 && $dmonth >= 6) { $dyear--; }
		$dyear = cdb_fns::getv("$activity year", $dyear);
	    } until ($dyear > 1990 && $dyear < 2100);
	    $daypermonth[2] = (isleapyear($dyear)) ? 29 : 28 ;
	    $mdays = $daypermonth[$dmonth];
	} until (cdb_fns::bounds($dday, 1, $mdays, "day")
	      && cdb_fns::ok());
	$ddate = pack_date($dyear, $dmonth, $dday);
	$fdate = format_date($ddate, "MM dd yyyy");
    }
    else {
	($dyear, $dmonth, $dday) = unpack("a4a2a2", $ddate);
    }
    print "\n\u\L$activity: $fdate\n";
    return($ddate, $dmonth, $dday, $dyear);
}

sub timestamp2date ($) {
my($date) = @_;
    $date =~ s/[^\d]//g;
    $date =~ s/(\d{8}).*/$1/;
    if($date eq "00000000") { $date = ""; }
    return($date);
}

# format date as mm/dd/yyyy
# if time is present, preserve that at end
# if optional second parameter (e.g. "mm yyyy") then format to that.
sub format_date ($;$) {
my($fdate, $format) = @_;
my(@months, $ftime);
    @months=("Jan", "Feb", "Mar", "Apr", "May", "Jun",
	     "Jul", "Aug", "Sep", "Oct", "Nov", "Dec");
    @longmonths=("January", "February", "March", "April", "May", "June",
	     "July", "August", "September", "October", "November", "December");
    $fdate ||= "";
    $format ||= "";
    ($fdate, $ftime) = split(' ', $fdate);
    $fdate ||= "";
    ($main::debug > 1) && print "format_date: fdate = \'$fdate\', ftime = \'$ftime\'\n";
    $fdate =~ s/-//g;
    if($format) {
	$fdate =~ /^(\d\d*)(\d\d)(\d\d$)/;
	$m = $2; $d = $3; $y = $1;
	$m ||= 0;
	$y ||= 0;
	$d ||= 0;
	$yy = sprintf("%02d", $y % 100);
	$format =~ s/mm*/$m/;
	$format =~ s/MM*/$months[$m-1]/;
	$format =~ s/LL*/$longmonths[$m-1]/;
	$format =~ s/dd/$d/;
	$format =~ s/yyyy/$y/i;
	$format =~ s/yy/$yy/;
	$fdate = $format;
    }
    else {
	$fdate =~ s/^(\d\d*)(\d\d)(\d\d)$/$2\/$3\/$1/;
    }
    if($ftime) { $fdate .= " $ftime"; }
    return($fdate);
}

# format date as mm/dd/yyyy
# if time is present, preserve that at end
# if optional second parameter (e.g. "mm yyyy") then format to that.
sub format_ts ($;$) {
my($fdate, $format) = @_;
my(@months, $ftime);
    @months=("Jan", "Feb", "Mar", "Apr", "May", "Jun",
	     "Jul", "Aug", "Sep", "Oct", "Nov", "Dec");
    $format ||= "";
    if($fdate =~ /(\d\d\d\d-*\d\d-*\d\d) *(\d\d:*\d\d:*\d\d)$/) {
	$fdate = $1;
	$ftime = $2;
    }
#   ($fdate, $ftime) = split(' ', $fdate);
    $main::debug && print "fdate = \'$fdate\', ftime = \'$ftime\'\n";
    $fdate =~ s/-//g;
    if($format) {
	$fdate =~ /^(\d\d*)(\d\d)(\d\d$)/;
	$m = $2; $d = $3; $y = $1;
	$yy = sprintf("%02d", $y % 100);
	$format =~ s/mm*/$m/;
	$format =~ s/MM*/$months[$m-1]/;
	$format =~ s/dd/$d/;
	$format =~ s/yyyy/$y/;
	$format =~ s/yy/$yy/;
	$fdate = $format;
    }
    else {
	$fdate =~ s/^(\d\d*)(\d\d)(\d\d)$/$2\/$3\/$1/;
    }
    if($ftime) { $fdate .= " $ftime"; }
    return($fdate);
}

=head1 add_time

=head2 DESCRIPTION

add time b to time a

=head2 ARGUMENTS

=over 5

=item string time a hhmmss or hh:mm:ss

=item string time b hhmmss or hh:mm:ss

=back

=head2 RETURN VALUES

=over 5

=item string time as hhmmss or hh:mm:ss, depending upon formatting of 'a'.  Can return an answer larger than 23:59:59.

=back

=cut
sub add_time ($$) {
my($a, $b) = @_;
my($ah, $am, $as, $bh, $bm, $bs, $r, $rh, $rm, $rs);
my($colon);
    if($a =~ /:/) { $colon = ":"; }
    ($ah, $am, $as) = unpack_time($a);
    ($bh, $bm, $bs) = unpack_time($b);
    $rs = $as + $bs;
    while($rs > 59) {
	$rs -= 60;
	$am++;
    }
    $rm = $am + $bm;
    while($rm > 59) {
	$rm -= 60;
	$ah++;
    }
    return(pack_time($ah + $bh, $rm, $rs, $colon));
}

=head1 add_time_c

=head2 DESCRIPTION

add time b to time a, return sum modulo 24:00:00 plus carry

=head2 ARGUMENTS

=over 5

=item string time a hhmmss or hh:mm:ss

=item string time b hhmmss or hh:mm:ss

=back

=head2 RETURN VALUES

=over 5

=item array of string time as hhmmss or hh:mm:ss, depending upon formatting of 'a', followed by carry value (integer hours divided by 24)

=back

=cut
sub add_time_c ($$) {
my($a, $b) = @_;
my($ah, $am, $as, $bh, $bm, $bs, $r, $rh, $rm, $rs);
my($colon, $carry);
    if($a =~ /:/) { $colon = ":"; }
    $carry = 0;
    ($ah, $am, $as) = unpack_time($a);
    ($bh, $bm, $bs) = unpack_time($b);
    $rs = $as + $bs;
    while($rs > 59) {
	$rs -= 60;
	$am++;
    }
    $rm = $am + $bm;
    while($rm > 59) {
	$rm -= 60;
	$ah++;
    }
    $carry = 0;
    $rh = $ah + $bh;
    while($rh > 23) {
	$carry++;
	$rh -= 24;
    }
    return(pack_time($rh, $rm, $rs, $colon), $carry);
}

=head1 sub_time_c

=head2 DESCRIPTION

subtract time b from time a, return answer modulo 24:00:00 plus borrow

=head2 ARGUMENTS

=over 5

=item string time a hhmmss or hh:mm:ss

=item string time b hhmmss or hh:mm:ss

=back

=head2 RETURN VALUES

=over 5

=item array of string time as hhmmss or hh:mm:ss, depending upon formatting of 'a', followed by borrow value (integer hours divided by 24)

=back

=cut
sub sub_time_c ($$) {
my($a, $b) = @_;
my($ah, $am, $as, $bh, $bm, $bs, $r, $rh, $rm, $rs);
my($colon, $borrow);
    if($a =~ /:/) { $colon = ":"; }
    ($ah, $am, $as) = unpack_time($a);
    ($bh, $bm, $bs) = unpack_time($b);
    $rs = $as - $bs;
    if($rs < 0) {
	$rs += 60;
	$am--;
    }
    $rm = $am - $bm;
    if($rm < 0) {
	$rm += 60;
	$ah--;
    }
    $rh = $ah - $bh;
    $borrow = 0;
    if($rh < 0) {
	$borrow = 1;
	$rh += 24;
    }
    return(pack_time($rh, $rm, $rs, $colon), $borrow);
}

=head1 sub_time

=head2 DESCRIPTION

subtract time b from time a

=head2 ARGUMENTS

=over 5

=item string time a hhmmss or hh:mm:ss

=item string time b hhmmss or hh:mm:ss

=back

=head2 RETURN VALUES

=over 5

=item array of string time as hhmmss or hh:mm:ss, depending upon formatting of 'a'

=back

=cut
sub sub_time ($$) {
my($a, $b) = @_;
my($ah, $am, $as, $bh, $bm, $bs, $r, $rh, $rm, $rs);
my($colon);
    if($a =~ /:/) { $colon = ":"; }
    ($ah, $am, $as) = unpack_time($a);
    ($bh, $bm, $bs) = unpack_time($b);
    $rs = $as - $bs;
    if($rs < 0) {
	$rs += 60;
	$am--;
    }
    $rm = $am - $bm;
    if($rm < 0) {
	$rm += 60;
	$ah--;
    }
    return(pack_time($ah - $bh, $rm, $rs, $colon));
}

=head1 unpack_time

=head2 DESCRIPTION

split time into hr, min, sec

=head2 ARGUMENTS

=over 5

=item string time as hhmmss or hh:mm:ss

=back

=head2 RETURN VALUES

=over 5

=item array of integers (hh, mm, ss)

=back

=cut
sub unpack_time ($) {
my($a) = @_;
my($ah, $am, $as);
    if($a =~ /(\d\d):*(\d\d):*(\d\d)/ || $a =~ /(\d+):(\d+):(\d+)/) {
	$ah = $1; $am = $2; $as = $3;
	$as = sprintf("%02d", $3);
	$am = sprintf("%02d", $2);
	$ah = sprintf("%02d", $1);
    }
    if($a =~ /^-/) {
	$ah = 0 - $ah;
    }
    return($ah, $am, $as);
}

=head1 unpack_timestamp

=head2 DESCRIPTION

split timestamp into date, time

=head2 ARGUMENTS

=over 5

=item timestamp yyyymmdd(hhmmss)?  Any punctuation is ignored, and stripped from the output.  Can be used to "clean up" dates or timestamps returned by mysql queries.  Also handles timestamp format "mm/dd/yyyy hh:mm:ss [AP]M"

=back

=head2 RETURN VALUES

=over 5

=item array of integers (yyyymmdd, hhmmss)

=back

=cut
sub unpack_timestamp ($) {
my($ts) = @_;
my($date, $time);
    if($ts =~ /(\d+)\/(\d+)\/(\d{4}) (\d{2}):(\d{2}):(\d{2}) ([AP])M/i) {
	$ts = sprintf("%4d%02d%02d%02d%02d%02d", $3, $1, $2, $4 + ($7 eq "P") ? 12 : 0, $5, $6);
    }
    $ts =~ s/[^\d]//g;
    ($date, $time) = unpack("a8a6", $ts);
    return($date, $time);
}

=head1 pack_time

=head2 DESCRIPTION

pack time into hh:mm:ss or hhmmss

=head2 ARGUMENTS

=over 5

=item integer hour 0-23

=item integer minute 0-59

=item integer second 0-59

=item integer optional flag, true if ':' should be used between fields

=back

=head2 RETURN VALUES

=over 5

=item string hhmmss or hh:mm:ss

=back

=cut
sub pack_time ($$$;$) {
my($ah, $am, $as, $colon) = @_;
my($r);
    if($colon) { $colon = ":"; }
    $as = sprintf("%02d", $as);
    $am = sprintf("%02d", $am);
    $ah = sprintf("%02d", $ah);
    $r = join($colon, ($ah, $am, $as));
    return($r);
}

sub Format_Time ($$$;$) {
my($h, $m, $s, $format)=@_;
my($t);
    $s ||= 0;
    $m ||= 0;
    $format ||= "\'%02d:%02d:%02d\'";
    $t = sprintf($format, $h, $m, $s);
    return($t);
}

sub get_activity_time ($$) {
my($ntime, $activity) = @_;
my($ftime, $nhour, $nmin, $nsec);

    $ntime ||= "000000";
    ($nhour, $nmin, $nsec) = unpack("a2a2a2", $ntime);
    $ftime = Format_Time($nhour, $nmin, $nsec);
    if(!cdb_fns::ok("Use this time ($ftime) for \u\L$activity time?")) {
	($nhour, $nmin, $nsec) = unpack("a2a2a2", $ntime);
	do {
	    do {
		$nhour=cdb_fns::getv("$activity hour (0 to 23)",$nhour);
	    } until (cdb_fns::bounds($nhour, 0, 23, "hour"));
	    do {
		$nmin = cdb_fns::getv("$activity minute (0 to 59):",$nmin);
	    } until (cdb_fns::bounds($nmin, 0, 59, "minute"));
	    do {
		$nsec = cdb_fns::getv("$activity second (0 to 59)", $nsec);
	    } until (cdb_fns::bounds($nsec, 0, 59, "second"));
	} until(cdb_fns::ok());
	$ntime = pack_time($nhour, $nmin, $nsec);
	$ftime = Format_Time($nhour, $nmin, $nsec);
    }
    print "\n\u\L$activity time: $ftime\n";
    return($ntime);
}

sub date2doy ($) {
my($date)=@_;
my($y, $m, $d, $doy);
    ($y, $m, $d) = unpack_date($date);
    unless($main::OS) {
	chomp($main::OS = `uname -s`);
    }
    if($main::OS eq "Linux") {
	chomp($doy = `/bin/date +%j -d ${y}${m}${d}`);
    }
    else {
	chomp($doy = `/bin/date -j +%j ${y}${m}${d}1200`);
    }
    return($doy);
}

sub julian2date ($) {
my($julian)=@_;
my($yr, $mon, $day, $date);
my($jd0, $z, $f, $a, $alp, $b, $c, $d, $e, $yc);

    $jd0 = $julian + 0.5;
    $z = int($jd0);
    $f = $jd0 - $z;
    $a = 0.0;
    $alp = 0.0;

    if ( $z < 2299161 ) {
	$a = $z;
    }
    else {
	$alp = int(($z - 1867216.25)/36524.25);
	$a = $z + 1.0 + $alp - int($alp/4.0);
    }

    $b = $a + 1524;
    $c = int(($b - 122.1)/365.25);
    $d = int(365.25 * $c);
    $e = int(($b - $d)/30.6001);

    $day = $b - $d - int(30.6001 * $e) + $f;

    $mon = 0;
    if ($e < 13.5) {
	$mon = $e - 1;
    }
    else {
	$mon = $e - 13;
    }

    $yr = 0;
    if ($mon > 2.5) {
	$yr = $c - 4716;
    }
    else {
	$yr = $c - 4715;
    }

    # $uth = int(24.0*($day - int($day)))
    # $utm = int(1440.0*($day - int($day) - $uth/24.0))
    # $uts = 86400.0*($day - int($day) - $uth/24.0 - $utm/1440.0)

    return(pack_date($yr, $mon, int($day)));
}

=head1 doy2date

=head2 DESCRIPTION

return date given day of the current year

=head2 ARGUMENTS

=over 5

=item integer 1-366

=back

=head2 RETURN VALUES

=over 5

=item date yyyymmdd

=back

=cut
sub doy2date ($) {
my($doy)=@_;
my($yc);
    $doy += 50448 + 2400000.5 ;
    unless($main::year > 1969) {
	($main::date,$main::month,$main::day,$main::year) = date();
    }
    for($yc = 1997 ; $yc < $main::year ; $yc++) {
	$doy += (isleapyear($yc)) ? 366 : 365 ;
    }
    return(julian2date($doy));
}

sub build_dow ($;$) {
my($y, $m)=@_;
my(@fields, $i, $ymd);
    if($y =~ /\d{6}/) {
	$ymd = $y;
	($y, $m, $i) = unpack_date($ymd);
    }
    else {
	$ymd = pack_date($y, $m);
    }
    # build array's of dates by day
    open(CAL, "cal $m $y |");
    while(<CAL>) {
	/$y/ && next;
	/Tu/ && next;
	chomp;
	@fields = unpack("a2 a3 a3 a3 a3 a3 a3", $_);
	for($i = 0 ; $i < 7 ; $i++) {
	    $fields[$i] += 0;
	    if($fields[$i]) {
		push(@ { $main::day2date{$daynames[$i]} }, $fields[$i]);
		$main::date2day{$fields[$i]} = $daynames[$i];
		$main::date2dno{$fields[$i]} = $i;
		$ { $main::month2date2day{$ymd} } {$fields[$i]} = $daynames[$i];
		$ { $main::month2date2dno{$ymd} } {$fields[$i]} = $i;
	    }
	}
    }
    close(CAL);
}

# sample output of cal 1 2003
#    January 2003
# S  M Tu  W Th  F  S
#          1  2  3  4
# 5  6  7  8  9 10 11
#12 13 14 15 16 17 18
#19 20 21 22 23 24 25
#26 27 28 29 30 31
#

=head1 get_dow

=head2 DESCRIPTION

return day of week 1-7 for today.

=head2 ARGUMENTS

=over 5

=item void

=back

=head2 RETURN VALUES

=over 5

=item integer 1-7, 1 = Monday

=back

=cut
sub get_dow () {
my($a);
    chomp($a = `/bin/date +%u`);
    return($a);
}

=head1 get_hour

=head2 DESCRIPTION

return hour 0-23 for now.

=head2 ARGUMENTS

=over 5

=item void

=back

=head2 RETURN VALUES

=over 5

=item integer 0-23

=back

=cut
sub get_hour () {
my($a);
    chomp($a = `/bin/date +%k`);
    return($a);
}

=head1 date2dow

=head2 DESCRIPTION

return day of week 0-6 given a date.

=head2 ARGUMENTS

=over 5

=item date yyyymmdd

=back

=head2 RETURN VALUES

=over 5

=item integer 0-6, 0 = Sunday

=back

=cut
sub date2dow ($) {
my($d)=@_;
my($dow);
    unless($main::OS) {
	chomp($main::OS = `uname -s`);
    }
    if($main::OS eq "Linux") {
	chomp($dow = `/bin/date +%u -d ${d}`);
    }
    else {
	chomp($dow = `/bin/date -j +%u ${d}1200`);
    }
    return($dow % 7);	# make 0 to 6, not 1 to 7
}

=head1 is_holiday

=head2 DESCRIPTION

see if date is a non-workday, and if so, return string description of reason.

=head2 ARGUMENTS

=over 5

=item date yyyymmdd

=item optional integer year yyyy

=item optional integer month mm

=item optional integer date dd

=back

=head2 BUGS

=item USA-centric idea of holidays

=back

=head2 RETURN VALUES

=over 5

=item string "" if business day, or "New Years Day", for example.

=back

=cut
sub is_holiday ($;$$$) {
my($rdate, $ry, $rm, $rd) = @_;
my($dno, $comment, $snoday, @snowdays, $snoreason, $pa);
    $rdate =~ s/-//g;
    if($rdate) {
	($ry, $rm, $rd) = unpack_date($rdate);
    }
    unless($rdate) { $rdate = pack_date($ry, $rm, $rd); }
    if(-s "/var/billing/snowdays") {
	chomp($snoday = `cat /var/billing/snowdays`);
	@snowdays = split(/\n/, $snoday);
	foreach $pa (@snowdays) {
	    ($snoday, $snoreason) = split(/\s/, $pa);
	    if($snoday == $rdate) { return($snoreason || "snow day"); }
	}
    }
    $dno = date2dow(pack_date($ry, $rm, $rd));
    unless($ry && $rm && $rd) { return(0); }
    $comment = "";
    if($dno == 6) { $comment = "Saturday"; }
    if($dno == 0 || $dno == 7) { $comment = "Sunday"; }
    if($rm == 12 && $rd == 31 && $dno == 5) { $comment = "Friday before New Years Day"; }
    if($rm ==  1 && $rd ==  1) { $comment = "New Years Day"; }
    if($rm ==  1 && $rd ==  2 && $dno == 1) { $comment = "Monday after New Years Day"; }
    if($rm ==  1 && $rd ==  2 && $dno == 5) { $comment = "Friday after New Years Day"; }
    if($rm ==  2 && $rd >= 15 && $rd <= 21 && $dno == 1) { $comment = "President's Day"; }
    if($rm ==  7 && $rd ==  3 && $dno == 5) { $comment = "Friday before Fourth of July"; }
    if($rm ==  7 && $rd ==  4) { $comment = "Fourth of July"; }
    if($rm ==  7 && $rd ==  5 && $dno == 1) { $comment = "Monday after Fourth of July"; }
    if($rm ==  5 && $rd >= 24 && $rd <= 30 && $dno == 1) { $comment = "Memorial Day"; }
    if($rm ==  9 && $rd <=  7 && $dno == 1) { $comment = "Labor Day"; }
    if($rm == 11 && $rd >= 22 && $rd <= 28 && $dno == 4) { $comment = "Thanksgiving"; }
    if($rm == 11 && $rd >= 23 && $rd <= 29 && $dno == 5) { $comment = "Friday after Thanksgiving"; }
    if($rm == 12 && $rd == 24) { $comment = "day before Christmas"; }
    if($rm == 12 && $rd == 25) { $comment = "Christmas"; }
    if($rm == 12 && $rd == 26 && $dno == 1) { $comment = "Monday after Christmas"; }
    if($rm == 12 && $rd == 26 && $dno == 5) { $comment = "Friday after Christmas"; }
    return($comment);
}

=head1 business_days_between

=head2 DESCRIPTION

compute the number of business days between two dates.

=head2 ARGUMENTS

=over 5

=item from_date yyyymmdd

=item to_date yyyymmdd

=back

=head2 RETURN VALUES

=over 5

=item integer number of business days from from_date to to_date

=back

=cut
sub business_days_between ($$) {
my($from_date, $to_date)=@_;
my($business_days, $rdate, $comment, $dno, $ry, $rm, $rd);
    $business_days = 0;
    for($rdate = $from_date ; $rdate < $to_date ; $rdate = n_days_hence($rdate, 1)) {
	$comment = is_holiday($rdate);
	unless($comment) { $business_days++; }
	if($main::debug) {
	    $dno = date2dow($rdate);
	    ($ry, $rm, $rd) = unpack_date($rdate);
	    $main::debug && printf("day %2d is %d, business_days = %d %s\n", $rd, $dno, $business_days, $comment);
	}
    }
    $main::debug && print "business days = $business_days between $from_date and $to_date";
    return($business_days);
}

=head1 sec_2_hms

=head2 DESCRIPTION

convert number representing seconds to hours:minutes:seconds string

=head2 ARGUMENTS

=over 5

=item seconds integer

=back

=head2 RETURN VALUES

=over 5

=item string hh:mm:ss

=back

=cut
sub sec_2_hms (*) {
my($sec)=@_;
my $h, $m, $s, $r;
    $s = $sec % 60;
    $sec = int($sec / 60);
    $m = $sec % 60;
    $h = int($sec / 60);
    return(sprintf("%d:%02d:%02d", $h, $m, $s));
}
