#!/usr/local/bin/perl
    # read backup file produced by Invoices & Estimates Pro
    # generate one line summary containing name,date,invnum,amount
    # order of one line summary is name first for 'sort by name'
    # order of one line summary is date first for 'sort by date'

    # datetime in invoice 516 (file 665) is 58520   4598.0000
    # invoice date is 2012 08 03 --> Julian date 2456142.5
    # Modified Julian date is 56142.0
    # unix time is 1343952000
    # 4598 represents days since 2000-01-01

    # datetime in invoice 507 (file 655) is 47071   4558.0000
    # invoice date is 2012 06 24 --> Julian date 2456102.5
    # Modified Julian date is 56102.0
    # unix time is 1340496000
    # 4558 represents days since 2000-01-01
    #
    # 2000-01-01 is julian day 2451544.5, mjd 51544
    #
    # these invoices are 40 days apart
    # magic time stamp difference 1: 58520-47071 = 11449
    # magic time stamp difference 2: 4598 - 4558 = 40
    # timedate value appears to be: seconds since midnight, days since 20000101

    use File::Basename;
    use lib dirname (__FILE__);
    use date_fns;

    $sort_by_name = 1;
    $sort_by_date = 0;
    my $invfile;
    for(my $i = 0 ; $i <= $#ARGV ; $i++) {
	if(-s $ARGV[$i]) {
	    $invfile = $ARGV[$i];
	}
	elsif($ARGV[$i] eq "-n") {
	    $sort_by_name = 1;
	    $sort_by_date = 0;
	}
	elsif($ARGV[$i] eq "-d") {
	    $sort_by_name = 0;
	    $sort_by_date = 1;
	}
	else {
	    goto Usage;
	}
    }
    unless(-s $invfile) {
Usage:	printf("Usage: generate_index.pl [-name] [-date] Items\\\\Invoice\\\\NNN.txt\n");
	exit(-1);
    }
    open(IN, "<$invfile");
    %fh = ();
    $lines = 0;
    while(<IN>) {
	chomp;
	s/
//;
	@fields = split(/\t/);
	# for(my $i = 0 ; $i <= $#fields ; $i++) {
	#   printf("%d	%s\n", $i, $fields[$i]);
	# }
	$fh{invnum} ||= $fields[0];
	if($fields[5] eq "text") {
	    if($fields[4] =~ /(^description)/) {
		$fh{$1 . $fields[3]} = $fields[6];
	    }
	    elsif($fields[6]
	    && $fields[6] !~ /^\s$/
	    && $fields[6] !~ /^\s*<.*>$/) {
		$fh{$fields[4]} = $fields[6];
	    }
	}
	elsif($fields[5] eq "number") {
	    if($fields[2] eq "master") {
		$fh{$fields[4]} = $fields[7];
	    }
	    else {
		$fh{$fields[4] . $fields[3]} = $fields[7];
	    }
	}
	elsif($fields[5] eq "datetime") {
	    $fh{$fields[4]} = Date_fns::format_date(Date_fns::n_days_hence(20000101, $fields[7]), "yyyy-mm-dd");
	    $fh{"raw_" . $fields[4]} = Date_fns::n_days_hence(20000101, $fields[7]);
	    my $k = $fields[4];
	    $k =~ s/date/time/;
	    $fh{$k} = Date_fns::sec_2_hms($fields[6]);
	}
	$lines++;
	# $lines && exit;
    }
    close(IN);
    $fh{'terms'} = int($fh{'terms'});
    $fh{'invoicenumber'} = int($fh{'invoicenumber'});
    if($fh{'raw_invoicedate'} && $fh{'terms'} > 0) {
	$fh{'duedate'} = Date_fns::format_date(Date_fns::n_days_hence($fh{'raw_invoicedate'}, $fh{'terms'}), "yyyy-mm-dd");
    }
    $fh{'total'} = 0 + $fh{'subtotal'} + $fh{'tabletax1'} + $fh{'shippingamount'};
    $fh{'balance_due'} = sprintf("\$%.2f", $fh{'total'} - $fh{'deposit'});
    $fh{'total'} = sprintf("\$%.2f", $fh{'total'});
    $fh{'deposit'} = sprintf("\$%.2f", 0 + $fh{'deposit'});
    if($debug) {
	foreach $key (sort keys %fh) {
	    printf("%s	%s\n", $key, $fh{$key});
	}
    }
    # printf("Invoice %d has Balance Due %s\n", $fh{'invoicenumber'}, $fh{'balance_due'});
    printf("%d\n", $fh{'invoicenumber'});
    if($sort_by_name) {
	open(TMPL, "<index_name.tmpl");
	open(OUT, ">>index_by_name.txt");
    }
    if($sort_by_date) {
	open(TMPL, "<index_date.tmpl");
	open(OUT, ">>index_by_date.txt");
    }
    my $tmpl;
    my $last;
    while(<TMPL>) {
	s/_(\w\w+)_/$fh{$1}/g;
	print OUT $_;
    }
    close(OUT);
    close(TMPL);

