#!/usr/local/bin/perl
    # read backup file produced by Invoices & Estimates Pro
    # generate troff file from template, use that to generate PDF

    # datetime in invoice 516 (file 665) is 58520   4598.0000
    # invoice date is 2012 08 03 --> Julian date 2456142.5
    # Modified Julian date is 56142.0
    # unix time is 1343952000
    # 4598 represents days since 2000-01-01

    # datetime in invoice 507 (file 655) is 47071   4558.0000
    # invoice date is 2012 06 24 --> Julian date 2456102.5
    # Modified Julian date is 56102.0
    # unix time is 1340496000
    # 4558 represents days since 2000-01-01
    #
    # 2000-01-01 is julian day 2451544.5, mjd 51544
    #
    # these invoices are 40 days apart
    # magic time stamp difference 1: 58520-47071 = 11449
    # magic time stamp difference 2: 4598 - 4558 = 40

    # timedate value appears to be: field[6]=seconds since midnight, field[7]=days since 20000101
    # four kinds of records:
#655		master	1	contact4	text	nobody@nowhere.com
#655		master	1	createddate	datetime	47071	4558.0000	
#655		master	1	partpayment	number		0
#655		activitytable	1	hidden	char(10)			
    # note: field 1 is empty
    # note: field 6 is empty for "number" records, field 7 has the value

    use File::Basename;
    use lib dirname (__FILE__);
    use date_fns;

    my $invfile;
    my $error = 0;
    my $debug_text = 0;
    my $debug_numbers = 0;

    for(my $i = 0 ; $i <= $#ARGV ; $i++) {
	if(-s $ARGV[$i]) {
	    $invfile = $ARGV[$i];
	}
	elsif($ARGV[$i] eq "-T") {
	    $debug_text = 1;
	}
	elsif($ARGV[$i] eq "-N") {
	    $debug_numbers = 1;
	}
	else {
	    $error = 1;
	    last;
	}
    }
    if($error || ! -s $invfile) {
Usage:	printf("Usage: recover_invoice.pl Items\\\\Invoice\\\\NNN.txt\n");
	exit(-1);
    }
    open(IN, "<$invfile");
    %fh = ();
    $lines = 0;
    while(<IN>) {
	chomp;
	s/
//;
	@fields = split(/\t/);
	# for(my $i = 0 ; $i <= $#fields ; $i++) {
	#   printf("%d	%s\n", $i, $fields[$i]);
	# }
	$fh{invnum} ||= $fields[0];
	if($fields[5] eq "text") {
	    if($fields[4] =~ /(^description)/) {
		$fh{$1 . $fields[3]} = $fields[6];
		$debug_text && printf("text description: fh{%s . %s} = %s\n", $1, $fields[3], $fields[6]);
	    }
	    elsif($fields[6]
	    && $fields[6] !~ /^\s$/
	    && $fields[6] !~ /^\s*<.*>$/) {
		$fh{$fields[4]} = $fields[6];
		$debug_text && printf("text other: fh{%s} = %s\n", $fields[4], $fields[6]);
	    }
	}
	elsif($fields[5] eq "number") {
	    if($fields[2] eq "master") {
		$fh{$fields[4]} = $fields[7];
	    }
	    else {
		$fh{$fields[4] . $fields[3]} = $fields[7];
	    }
	}
	elsif($fields[5] eq "datetime") {
	    $fh{$fields[4]} = Date_fns::format_date(Date_fns::n_days_hence(20000101, $fields[7]), "mm/dd/yyyy");
	    $fh{"raw_" . $fields[4]} = Date_fns::n_days_hence(20000101, $fields[7]);
	    my $k = $fields[4];
	    $k =~ s/date/time/;
	    $fh{$k} = Date_fns::sec_2_hms($fields[6]);
	}
	$lines++;
	# $lines && exit;
    }
    close(IN);
    $fh{'terms'} = int($fh{'terms'});
    $fh{'invoicenumber'} = int($fh{'invoicenumber'});
    if($fh{'raw_invoicedate'} && $fh{'terms'} > 0) {
	$fh{'duedate'} = Date_fns::format_date(Date_fns::n_days_hence($fh{'raw_invoicedate'}, $fh{'terms'}), "mm/dd/yyyy");
    }
    $fh{'total'} = 0 + $fh{'subtotal'} + $fh{'tabletax1'} + $fh{'shippingamount'};
    $fh{'balance_due'} = sprintf("\$%.2f", $fh{'total'} - $fh{'deposit'});
    $fh{'total'} = sprintf("\$%.2f", $fh{'total'});
    $fh{'subtotal'} = sprintf("\$%.2f", 0 + $fh{'subtotal'});
    $fh{'tabletax1'} = sprintf("\$%.2f", $fh{'tabletax1'});
    $fh{'tabletax2'} = sprintf("\$%.2f", $fh{'tabletax2'});
    $fh{'taxrate1'} = sprintf("%.2f", $fh{'taxrate1'});
    $fh{'shippingamount'} = sprintf("\$%.2f", 0 + $fh{'shippingamount'});
    $fh{'deposit'} = sprintf("\$%.2f", 0 + $fh{'deposit'});
    if($debug) {
	foreach $key (sort keys %fh) {
	    printf("%s	%s\n", $key, $fh{$key});
	}
    }
    # printf("Invoice %d has Balance Due %s\n", $fh{'invoicenumber'}, $fh{'balance_due'});
    printf("%d\n", $fh{'invoicenumber'});
    open(TMPL, "<invoice.nr");
    open(OUT, ">invoice.$fh{invoicenumber}.nr");
    my $tmpl;
    my $last;
    while(<TMPL>) {
	if(/_REPEAT_/) {
	    do {
		$last = <TMPL>;
		if($last !~ /_END_REPEAT_/) {
		    $tmpl .= $last;
		}
	    } until(!$last || $last =~ /_END_REPEAT_/);
	    for(my $i = 1; $fh{"description" . $i} ; $i++) {
		if($fh{"description" . $i}) {
		    $_ = $tmpl;
		    $_ =~ s/code/code$i/;
		    $_ =~ s/description/description$i/;
		    $_ =~ s/qty/qty$i/;
		    $_ =~ s/rate/rate$i/;
		    $fh{"qty$i"} = sprintf("%.2f", $fh{"qty$i"});
		    $fh{"rate$i"} = sprintf("%.2f", $fh{"rate$i"});
		    $fh{"amount"} = sprintf("%.2f", $fh{"rate$i"} * $fh{"qty$i"});
		    s/_(\w\w+)_/$fh{$1}/g;
		    print OUT $_;
		}
	    }
	}
	else {
	    s/_(\w[_\w]+)_/$fh{$1}/g;
	    print OUT $_;
	}
    }

